@@ -43,11 +43,11 @@ import ansible.errors
 import ansible.plugins.connection
 
 import mitogen.core
+import mitogen.parent
+import mitogen.service
 
 import ansible_mitogen.mixins
-import ansible_mitogen.parsing
 import ansible_mitogen.process
-import ansible_mitogen.services
 import ansible_mitogen.target
 import ansible_mitogen.transport_config
 import ansible_mitogen.utils.unsafe
@@ -145,9 +145,9 @@ def _connect_ssh(spec):
             'identity_file': private_key_file,
             'identities_only': False,
             'ssh_path': spec.ssh_executable(),
-            'connect_timeout': spec.ansible_ssh_timeout(),
+            'connect_timeout': spec.timeout(),
             'ssh_args': spec.ssh_args(),
-            'ssh_debug_level': spec.mitogen_ssh_debug_level(),
+            'ssh_debug_level': spec.verbosity(),
             'remote_name': get_remote_name(spec),
             'keepalive_count': (
                 spec.mitogen_ssh_keepalive_count() or 10
@@ -169,7 +169,7 @@ def _connect_buildah(spec):
             'username': spec.remote_user(),
             'container': spec.remote_addr(),
             'python_path': spec.python_path(),
-            'connect_timeout': spec.ansible_ssh_timeout() or spec.timeout(),
+            'connect_timeout': spec.timeout(),
             'remote_name': get_remote_name(spec),
         }
     }
@@ -185,7 +185,7 @@ def _connect_docker(spec):
             'username': spec.remote_user(),
             'container': spec.remote_addr(),
             'python_path': spec.python_path(rediscover_python=True),
-            'connect_timeout': spec.ansible_ssh_timeout() or spec.timeout(),
+            'connect_timeout': spec.timeout(),
             'remote_name': get_remote_name(spec),
         }
     }
@@ -200,7 +200,7 @@ def _connect_kubectl(spec):
         'kwargs': {
             'pod': spec.remote_addr(),
             'python_path': spec.python_path(),
-            'connect_timeout': spec.ansible_ssh_timeout() or spec.timeout(),
+            'connect_timeout': spec.timeout(),
             'kubectl_path': spec.mitogen_kubectl_path(),
             'kubectl_args': spec.extra_args(),
             'remote_name': get_remote_name(spec),
@@ -218,7 +218,7 @@ def _connect_jail(spec):
             'username': spec.remote_user(),
             'container': spec.remote_addr(),
             'python_path': spec.python_path(),
-            'connect_timeout': spec.ansible_ssh_timeout() or spec.timeout(),
+            'connect_timeout': spec.timeout(),
             'remote_name': get_remote_name(spec),
         }
     }
@@ -234,7 +234,7 @@ def _connect_lxc(spec):
             'container': spec.remote_addr(),
             'python_path': spec.python_path(),
             'lxc_attach_path': spec.mitogen_lxc_attach_path(),
-            'connect_timeout': spec.ansible_ssh_timeout() or spec.timeout(),
+            'connect_timeout': spec.timeout(),
             'remote_name': get_remote_name(spec),
         }
     }
@@ -250,7 +250,7 @@ def _connect_lxd(spec):
             'container': spec.remote_addr(),
             'python_path': spec.python_path(),
             'lxc_path': spec.mitogen_lxc_path(),
-            'connect_timeout': spec.ansible_ssh_timeout() or spec.timeout(),
+            'connect_timeout': spec.timeout(),
             'remote_name': get_remote_name(spec),
         }
     }
@@ -273,7 +273,7 @@ def _connect_podman(spec):
             'username': spec.remote_user(),
             'container': spec.remote_addr(),
             'python_path': spec.python_path(rediscover_python=True),
-            'connect_timeout': spec.ansible_ssh_timeout() or spec.timeout(),
+            'connect_timeout': spec.timeout(),
             'remote_name': get_remote_name(spec),
         }
     }
@@ -409,6 +409,7 @@ def _connect_mitogen_doas(spec):
 #: generating ContextService keyword arguments matching a connection
 #: specification.
 CONNECTION_METHOD = {
+    # Ansible connection plugins
     'buildah': _connect_buildah,
     'docker': _connect_docker,
     'kubectl': _connect_kubectl,
@@ -421,9 +422,14 @@ CONNECTION_METHOD = {
     'setns': _connect_setns,
     'ssh': _connect_ssh,
     'smart': _connect_ssh,  # issue #548.
+
+    # Ansible become plugins
+    'community.general.doas': _connect_doas,
     'su': _connect_su,
     'sudo': _connect_sudo,
     'doas': _connect_doas,
+
+    # Mitogen specific methods
     'mitogen_su': _connect_mitogen_su,
     'mitogen_sudo': _connect_mitogen_sudo,
     'mitogen_doas': _connect_mitogen_doas,
@@ -767,7 +773,7 @@ class Connection(ansible.plugins.connection.ConnectionBase):
                               C.BECOME_ALLOW_SAME_USER):
             stack += (CONNECTION_METHOD[spec.become_method()](spec),)
 
-        return stack
+        return ansible_mitogen.utils.unsafe.cast(stack)
 
     def _build_stack(self):
         """
@@ -890,6 +896,29 @@ class Connection(ansible.plugins.connection.ConnectionBase):
             self.binding.close()
             self.binding = None
 
+    def _mitogen_var_options(self, templar):
+        # Workaround for https://github.com/ansible/ansible/issues/84238
+        var_names = C.config.get_plugin_vars('connection', self._load_name)
+        variables = templar.available_variables
+        var_options = {
+            var_name: templar.template(variables[var_name])
+            for var_name in var_names
+            if var_name in variables
+        }
+
+        if self.allow_extras:
+            extras_var_prefix = 'ansible_%s_' % self.extras_prefix
+            var_options['_extras'] = {
+                var_name: templar.template(variables[var_name])
+                for var_name in variables
+                if var_name not in var_options
+                and var_name.startswith(extras_var_prefix)
+            }
+        else:
+            var_options['_extras'] = {}
+
+        return var_options
+
     reset_compat_msg = (
         'Mitogen only supports "reset_connection" on Ansible 2.5.6 or later'
     )
@@ -910,17 +939,38 @@ class Connection(ansible.plugins.connection.ConnectionBase):
                 self.reset_compat_msg
             )
 
-        # Strategy's _execute_meta doesn't have an action obj but we'll need one for
-        # running interpreter_discovery
-        # will create a new temporary action obj for this purpose
-        self._action = ansible_mitogen.mixins.ActionModuleMixin(
-            task=0,
-            connection=self,
-            play_context=self._play_context,
-            loader=0,
-            templar=0,
-            shared_loader_obj=0
-        )
+        # Handle templated connection variables during `meta: reset_connection`.
+        # Many bugs/implementation details of Mitogen & Ansible collide here.
+        # See #1079, #1096, #1132, ansible/ansible#84238, ...
+        try:
+            task, templar = self._play_context.vars.pop(
+                '_mitogen.smuggled.reset_connection',
+            )
+        except KeyError:
+            self._action_monkey_patched_by_mitogen = False
+        else:
+            # LOG.info('%r.reset(): remote_addr=%r', self, self._play_context.remote_addr)
+            # ansible.plugins.strategy.StrategyBase._execute_meta() doesn't
+            # have an action object, which we need for interpreter_discovery.
+            # Create a temporary action object for this purpose.
+            self._action = ansible_mitogen.mixins.ActionModuleMixin(
+                task=task,
+                connection=self,
+                play_context=self._play_context,
+                loader=templar._loader,
+                templar=templar,
+                shared_loader_obj=0,
+            )
+            self._action_monkey_patched_by_mitogen = True
+
+            # Workaround for https://github.com/ansible/ansible/issues/84238
+            self.set_options(
+                task_keys=task.dump_attrs(),
+                var_options=self._mitogen_var_options(templar),
+            )
+
+            del task
+            del templar
 
         # Clear out state in case we were ever connected.
         self.close()
@@ -941,6 +991,11 @@ class Connection(ansible.plugins.connection.ConnectionBase):
         finally:
             binding.close()
 
+        # Cleanup any monkey patching we did for `meta: reset_connection`
+        if self._action_monkey_patched_by_mitogen:
+            del self._action
+        del self._action_monkey_patched_by_mitogen
+
     # Compatibility with Ansible 2.4 wait_for_connection plug-in.
     _reset = reset
 
